from typing import Dict, Optional, Set

from tree_sitter import Node

from castep_linter.fortran.fortran_nodes.argument_types import ArgumentItem
from castep_linter.fortran.fortran_nodes.fortran_argument_list import FortranArgumentList
from castep_linter.fortran.fortran_nodes.fortran_node import FortranNode
from castep_linter.fortran.fortran_raw_types import Fortran, FType
from castep_linter.fortran.identifier import Identifier
from castep_linter.fortran.node_type_err import WrongNodeError


class FortranVariableDeclaration(FortranNode):
    """Class representing a variable declaration"""

    def __init__(self, var_decl_node: Node) -> None:
        super().__init__(var_decl_node)

        self.var_type = self.parse_fort_type()
        self.qualifiers = self.parse_fort_type_qualifiers()
        self.vars = self.parse_fort_var_names()
        self.args = self.parse_fort_var_size()

    def get_arg(self, keyword: Identifier, position: Optional[int] = None) -> ArgumentItem:
        """Get an argument from the call expression"""
        return self.args.get_arg(keyword, position)

    def get_initialized_vars(self) -> dict[Identifier, str]:
        """Get variables that have an initial value"""
        return {k: v for k, v in self.vars.items() if v}

    def parse_fort_type(self) -> FType:
        """Parse a variable declaration for type"""
        try:
            fortran_type = self.get(Fortran.INTRINSIC_TYPE).raw.upper()
            if fortran_type == "DOUBLE PRECISION":
                return FType.DOUBLE
            else:
                return FType[fortran_type]
        except KeyError:
            return FType.OTHER

    def parse_fort_type_qualifiers(self) -> Set[str]:
        """Parse a variable declaration for qualifiers, eg parameter"""
        qualifiers = set()
        for type_qualifier in self.get_children_by_name(Fortran.TYPE_QUALIFIER):
            qualifier = type_qualifier.raw.lower()
            qualifiers.add(qualifier)
        return qualifiers

    def parse_fort_var_size(self) -> FortranArgumentList:
        """Parse a variable declaration for a size, eg kind=8"""
        try:
            fortran_size = self.get(Fortran.SIZE).get(Fortran.ARGUMENT_LIST)
        except KeyError:
            return FortranArgumentList(None)

        if not isinstance(fortran_size, FortranArgumentList):
            err = f"Expected an argument list but got {fortran_size}"
            raise WrongNodeError(err)

        return fortran_size

    def parse_fort_var_names(self) -> Dict[Identifier, Optional[str]]:
        """Parse variable declaration statement for variables and optionally assignments"""
        myvars: Dict[Identifier, Optional[str]] = {}
        for assignment in self.get_children_by_name(Fortran.ASSIGNMENT_STMT):
            lhs, rhs = assignment.split()
            #   lhs, rhs = split_relational_node(assignment)
            varname = Identifier.from_node(lhs)
            if rhs.is_type(Fortran.STRING_LITERAL):
                myvars[varname] = rhs.parse_string_literal()
            else:
                myvars[varname] = None
        return myvars



"""Test that all real values are specified by real(kind=dp)"""

from castep_linter.error_logging import ErrorLogger
from castep_linter.fortran.fortran_nodes import (
    FortranNode,
    FortranVariableDeclaration,
)
from castep_linter.fortran.fortran_nodes.argument_types import PositionalArgument
from castep_linter.fortran.fortran_raw_types import Fortran, FType
from castep_linter.fortran.identifier import Identifier
from castep_linter.fortran.node_type_err import WrongNodeError
from castep_linter.tests import castep_identifiers


def check_real_dp_declaration(node: FortranNode, error_log: ErrorLogger) -> None:
    """Test that all real values are specified by real(kind=dp)"""

    if not isinstance(node, FortranVariableDeclaration):
        err = "Expected variable declaration node"
        raise WrongNodeError(err)

    if node.var_type not in [FType.REAL, FType.COMPLEX]:
        return

    try:
        arg = node.get_arg(position=1, keyword=Identifier("kind"))
    except KeyError:
        error_log.add_msg("Error", node, "No kind specifier")
        return

    if arg.value.ftype == Fortran.NUMBER_LITERAL:
        error_log.add_msg("Error", arg.value, "Numeric kind specifier")

    elif (
        arg.value.ftype == Fortran.IDENTIFIER
        and Identifier.from_node(arg.value) not in castep_identifiers.DP_ALL
    ):
        error_log.add_msg("Warning", arg.value, "Invalid kind specifier")

    elif isinstance(arg, PositionalArgument):
        error_log.add_msg("Info", arg.value, "Kind specified without keyword")



"""Fortran identifier Class Module"""

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from castep_linter.fortran.fortran_nodes import FortranNode


class Identifier:
    """Represents a type insensitive identifier in Fortran"""

    def __init__(self, name: str):
        self.lower_name = name.lower()

    @staticmethod
    def from_node(node: "FortranNode") -> "Identifier":
        """Create an identifier directly from a node"""
        return Identifier(node.raw)

    def __repr__(self):
        return self.lower_name

    def __hash__(self):
        return hash(self.lower_name)

    def __eq__(self, other):
        if other is None:
            return False
        if isinstance(other, str):
            return self.lower_name == other.lower()
        if isinstance(other, Identifier):
            return self.lower_name == other.lower_name.lower()
        err = "Can only compare identifiers with other identifiers"
        raise TypeError(err)



from castep_linter.fortran.fortran_nodes import FortranNode
from castep_linter.fortran.fortran_raw_types import Fortran
from castep_linter.fortran.identifier import Identifier


class FortranSubroutineNode(FortranNode):
    """Node representing a subroutine"""

    def get_context_identifier(self) -> Identifier:
        """Get the name of the subroutine"""
        return Identifier.from_node(self.get(Fortran.SUBROUTINE_STMT).get(Fortran.NAME))



"""Test that a call of complex(x) has a dp"""

from castep_linter.error_logging import ErrorLogger
from castep_linter.fortran.fortran_nodes import FortranCallExpression, FortranNode
from castep_linter.fortran.node_type_err import WrongNodeError
from castep_linter.tests import castep_identifiers


def check_complex_has_dp(node: FortranNode, error_log: ErrorLogger) -> None:
    """Test that a call of complex(x) has a dp"""

    if not isinstance(node, FortranCallExpression):
        err = "Expected variable declaration node"
        raise WrongNodeError(err)

    if node.name == castep_identifiers.CMPLX:
        try:
            arg_value = node.get_arg(position=3, keyword=castep_identifiers.KIND).value
        except KeyError:
            error_log.add_msg("Error", node, "No kind specifier in complex intrinsic")
            return

        if arg_value.raw.lower() not in castep_identifiers.DP_ALL:
            error_log.add_msg("Error", node, "Invalid kind specifier in complex intrinsic")



"""Test that allocate stat is used and checked"""

import logging

from castep_linter.error_logging import ErrorLogger
from castep_linter.fortran.fortran_nodes import (
    FortranCallExpression,
    FortranNode,
)
from castep_linter.fortran.fortran_raw_types import Fortran
from castep_linter.fortran.identifier import Identifier
from castep_linter.fortran.node_type_err import WrongNodeError
from castep_linter.tests import castep_identifiers


def check_allocate_error_names(
    node: FortranNode,
    error_log: ErrorLogger,
    variable_name: Identifier,
    subroutine_name: Identifier,
):
    """Check that an allocation error is correctly handled"""
    if not node.is_type(Fortran.IF_STMT):
        err = f"Expected if statement but got {node.type}"
        raise WrongNodeError(err)

    sub_node = node.get(Fortran.SUBROUTINE_CALL)
    if not isinstance(sub_node, FortranCallExpression):
        err = f"Expected call expression but got {node.type}"
        raise WrongNodeError(err)

    sub_call = sub_node

    if sub_call.name not in castep_identifiers.IO_ABORT_ANY:
        error_log.add_msg("Error", sub_node, "Allocation failure does not raise an error")
        return

    # Only do extra checking on io_allocate_abort -- ignore old io_abort style errors
    if sub_call.name != castep_identifiers.IO_ALLOCATE_ABORT:
        return

    error_array_node = sub_call.get_arg(position=1, keyword=castep_identifiers.IO_AA_ARRAY).value
    error_routine_node = sub_call.get_arg(
        position=2, keyword=castep_identifiers.IO_AA_ROUTINE
    ).value

    try:
        error_array = Identifier(error_array_node.parse_string_literal())
    except WrongNodeError:
        logging.warning("Failed to properly parse array name from: %s", node.raw)
        error_array = None

    try:
        error_routine = Identifier(error_routine_node.parse_string_literal())
    except WrongNodeError:
        logging.warning("Failed to properly parse subroutine name from: %s", node.raw)
        error_routine = None

    if not error_array == variable_name:
        err = f"Wrong array name used in io_allocate_abort. Expected: {variable_name}"
        error_log.add_msg("Error", error_array_node, err)

    if not error_routine == subroutine_name:
        err = f"Wrong subroutine name used in io_allocate_abort. Expected: {subroutine_name}"
        error_log.add_msg("Error", error_routine_node, err)


def variable_name_from_allocate(routine: FortranCallExpression):
    """Get the array from allocate statement"""
    allocate_var_node = routine.get_arg(position=1, keyword=castep_identifiers.ARRAY).value

    if routine.node.type == "call_expression":
        try:
            array_name_node = allocate_var_node.children[0]
        except IndexError:
            return ""
    else:
        array_name_node = allocate_var_node.get(Fortran.IDENTIFIER)

    try:
        return Identifier.from_node(array_name_node)
    except KeyError as exc:
        msg = "Could not identify variable name in allocate statement"
        routine.print_tree()
        raise ValueError(msg) from exc


def check_allocate_has_stat(node: FortranNode, error_log: ErrorLogger) -> None:
    """Test that allocate stat is used and checked"""

    if not isinstance(node, FortranCallExpression):
        err = f"Expected call expression but got {node.type}"
        raise WrongNodeError(err)

    routine = node

    if routine.name is None:
        return

    # Check this is actually an allocate statement
    if routine.name != castep_identifiers.ALLOCATE:
        return

    # First get the stat variable for this allocate statement
    try:
        stat_variable_node = routine.get_arg(keyword=castep_identifiers.STAT).value
    except KeyError:
        error_log.add_msg("Warning", node, "No stat on allocate statement")
        return

    # allocate_var_identifier = variable_name_from_allocate(routine)

    stat_variable = Identifier.from_node(stat_variable_node)

    # Find the next non-comment line
    next_node = node.next_named_sibling()
    while next_node and next_node.is_type(Fortran.COMMENT):
        next_node = next_node.next_named_sibling()

    # Check if that uses the stat variable
    stat_checked = False
    if next_node and next_node.is_type(Fortran.IF_STMT):
        try:
            relational_expr = next_node.get(Fortran.PAREN_EXPRESSION).get(Fortran.RELATIONAL_EXPR)
        except KeyError:
            error_log.add_msg("Error", stat_variable_node, "Allocate status not checked")
            return

        lhs, rhs = relational_expr.split()

        if lhs.is_type(Fortran.IDENTIFIER) and Identifier.from_node(lhs) == stat_variable:
            stat_checked = True

        if rhs.is_type(Fortran.IDENTIFIER) and Identifier.from_node(rhs) == stat_variable:
            stat_checked = True

    if not next_node or not stat_checked:
        error_log.add_msg("Error", stat_variable_node, "Allocate status not checked")
        return

    # context_name = node.get_context_identifier()
    # check_allocate_error_names(next_node, error_log, allocate_var_identifier, context_name)



# pylint: disable=W0621,C0116,C0114
from unittest import mock

import pytest

from castep_linter.fortran.node_type_err import WrongNodeError
from castep_linter.scan_files import run_tests_on_code
from castep_linter.tests import CheckFunctionDict, check_real_dp_declaration
from tests.conftest import Parser


@pytest.fixture
def test_list() -> CheckFunctionDict:
    return {"variable_declaration": [check_real_dp_declaration]}


def test_wrong_node():
    mock_node = mock.Mock(**{"is_type.return_value": False})
    err_log = mock.MagicMock()
    with pytest.raises(WrongNodeError):
        check_real_dp_declaration(mock_node, err_log)


def test_real_dp_correct(parse: Parser, test_list: CheckFunctionDict):
    code = b"real(kind=dp) :: y"
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 0


def test_real_version_kind_correct(parse: Parser, test_list: CheckFunctionDict):
    code = b"real(kind=version_kind) :: y"
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 0


def test_real_dp_by_position(parse: Parser, test_list: CheckFunctionDict):
    code = b"real(dp) :: y"
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 1


def test_real_dp_by_d0(parse: Parser, test_list: CheckFunctionDict):
    code = b"DOUBLE PRECISION :: y"
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 0


def test_real_dp_no_kind(parse: Parser, test_list: CheckFunctionDict):
    code = b"real, intent(in) :: y"
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 1


def test_real_integer_kind_with_keyword(parse: Parser, test_list: CheckFunctionDict):
    code = b"real(kind=8) :: y"
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 1


def test_real_integer_kind_without_keyword(parse: Parser, test_list: CheckFunctionDict):
    code = b"real(8) :: y"
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 1


def test_real_bad_var_kind_without_keyword(parse: Parser, test_list: CheckFunctionDict):
    code = b"real(x) :: y"
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 1


def test_real_bad_var_kind_with_keyword(parse: Parser, test_list: CheckFunctionDict):
    code = b"real(kind=x) :: y"
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 1


def test_real_with_other_keyword(parse: Parser, test_list: CheckFunctionDict):
    code = b"real(lemon=x) :: y"
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 1


def test_integer_declaration(parse: Parser, test_list: CheckFunctionDict):
    code = b"integer :: y"
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 0


def test_derived_type_declaration(parse: Parser, test_list: CheckFunctionDict):
    code = b"type(z) :: y"
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 0



"""General fixtures for testing castep-linter"""

from typing import Callable

import pytest

from castep_linter.fortran.parser import FortranTree

Parser = Callable[[bytes], FortranTree]


@pytest.fixture
def parse() -> Parser:
    """Get a general code parser"""

    def _parse(code: bytes) -> FortranTree:
        return FortranTree(code)

    return _parse


CodeWrapper = Callable[[bytes], bytes]


@pytest.fixture
def subroutine_wrapper() -> CodeWrapper:
    """Wrapper to give a subroutine context to code"""

    def _subroutine_wrapper(code: bytes) -> bytes:
        return (
            b"""module foo
            subroutine x(y)
            """
            + code
            + b"""
            end subroutine x
            end module foo"""
        )

    return _subroutine_wrapper


@pytest.fixture
def function_wrapper() -> CodeWrapper:
    """Wrapper to give a function context to code"""

    def _function_wrapper(code: bytes) -> bytes:
        return (
            b"""module foo
            function x(y)
            """
            + code
            + b"""
            end function x
            end module foo"""
        )

    return _function_wrapper



# pylint: disable=W0621,C0116,C0114
import pytest

from castep_linter.scan_files import run_tests_on_code
from castep_linter.tests import CheckFunctionDict, check_trace_entry_exit
from tests.conftest import CodeWrapper, Parser


@pytest.fixture
def test_list() -> CheckFunctionDict:
    return {"subroutine": [check_trace_entry_exit], "function": [check_trace_entry_exit]}


def test_trace_entry_exit_correct(
    parse: Parser, subroutine_wrapper: CodeWrapper, test_list: CheckFunctionDict
):
    code = subroutine_wrapper(
        b"""
    call trace_entry("x", stat)
    call trace_exit("x", stat)
    """
    )
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 0, error_log.errors


def test_trace_entry_exit_correct_extra(
    parse: Parser, subroutine_wrapper: CodeWrapper, test_list: CheckFunctionDict
):
    code = subroutine_wrapper(
        b"""
    call trace_entry("x", stat)
    call bleh()
    call trace_exit("x", stat)
    """
    )
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 0


def test_trace_entry_exit_correct_keyword(
    parse: Parser, subroutine_wrapper: CodeWrapper, test_list: CheckFunctionDict
):
    code = subroutine_wrapper(
        b"""
    call trace_entry(string="x", status=stat)
    call trace_exit(string="x", status=stat)
    """
    )
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 0


def test_trace_entry_exit_correct_by_param(
    parse: Parser, subroutine_wrapper: CodeWrapper, test_list: CheckFunctionDict
):
    code = subroutine_wrapper(
        b"""
    character(len=100), parameter :: sub_name = "x"
    call trace_entry(sub_name, stat)
    call trace_exit(sub_name, stat)
    """
    )
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 0


def test_trace_entry_exit_correct_by_param_extra(
    parse: Parser, subroutine_wrapper: CodeWrapper, test_list: CheckFunctionDict
):
    code = subroutine_wrapper(
        b"""
    character(len=100), parameter :: sub_name = "x"
    character(len=100), parameter :: bleh = othervar
    integer :: p
    type(myvar) :: z
    call trace_entry(sub_name, stat)
    call trace_exit(sub_name, stat)
    """
    )
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 0


def test_trace_entry_missing(
    parse: Parser, subroutine_wrapper: CodeWrapper, test_list: CheckFunctionDict
):
    code = subroutine_wrapper(b"""call trace_exit("x", stat)""")
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 1


def test_trace_exit_missing(
    parse: Parser, subroutine_wrapper: CodeWrapper, test_list: CheckFunctionDict
):
    code = subroutine_wrapper(b"""call trace_entry("x", stat)""")
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 1


def test_trace_entry_exit_missing(
    parse: Parser, subroutine_wrapper: CodeWrapper, test_list: CheckFunctionDict
):
    code = subroutine_wrapper(b"")
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 2


def test_trace_entry_exit_wrong_name(
    parse: Parser, subroutine_wrapper: CodeWrapper, test_list: CheckFunctionDict
):
    code = subroutine_wrapper(
        b"""
    call trace_entry("y", stat)
    call trace_exit("y", stat)
    """
    )
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 2


def test_trace_entry_exit_wrong_name_keyword(
    parse: Parser, subroutine_wrapper: CodeWrapper, test_list: CheckFunctionDict
):
    code = subroutine_wrapper(
        b"""
    call trace_entry(string="y", status=stat)
    call trace_exit(string="y", status=stat)
    """
    )
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 2


def test_trace_entry_exit_wrong_name_by_param(
    parse: Parser, subroutine_wrapper: CodeWrapper, test_list: CheckFunctionDict
):
    code = subroutine_wrapper(
        b"""
    character(len=100), parameter :: sub_name = "y"
    call trace_entry(sub_name, stat)
    call trace_exit(sub_name, stat)
    """
    )
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 2


def test_trace_entry_exit_no_name(
    parse: Parser, subroutine_wrapper: CodeWrapper, test_list: CheckFunctionDict
):
    code = subroutine_wrapper(
        b"""
    call trace_entry()
    call trace_exit()
    """
    )
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 2


def test_trace_entry_exit_unknown_name(
    parse: Parser, subroutine_wrapper: CodeWrapper, test_list: CheckFunctionDict
):
    code = subroutine_wrapper(
        b"""
    call trace_entry(other_var)
    call trace_exit(other_var)
    """
    )
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 2


def test_trace_entry_exit_correct_caps(parse: Parser, test_list: CheckFunctionDict):
    code = b"""
    module foo
    SUBROUTINE X(Y)
    CALL TRACE_ENTRY("X", STAT)
    CALL TRACE_EXIT("X", STAT)
    END SUBROUTINE X
    end module foo
    """
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 0, error_log.errors[0].message


def test_trace_entry_exit_correct_by_param_mixed_caps(parse: Parser, test_list: CheckFunctionDict):
    code = b"""
    module foo
    SUBROUTINE X(Y)
    CHARACTER(len=100), PARAMETER :: sub_name = "X"
    CALL TRACE_ENTRY(SUB_NAME, STAT)
    CALL TRACE_EXIT(SUB_NAME, STAT)
    END SUBROUTINE X
    end module foo
    """
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 0


def test_trace_entry_exit_correct_by_param_all_caps(parse: Parser, test_list: CheckFunctionDict):
    code = b"""
    module foo
    SUBROUTINE X(Y)
    CHARACTER(len=100), PARAMETER :: SUB_NAME = "X"
    CALL TRACE_ENTRY(SUB_NAME, STAT)
    CALL TRACE_EXIT(SUB_NAME, STAT)
    END SUBROUTINE X
    end module foo
    """
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 0


def test_trace_entry_exit_correct_function(
    parse: Parser, function_wrapper: CodeWrapper, test_list: CheckFunctionDict
):
    code = function_wrapper(
        b"""
        CALL TRACE_ENTRY("x", STAT)
        CALL TRACE_EXIT("x", STAT)
    """
    )
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 0


def test_trace_entry_exit_missing_function(
    parse: Parser, function_wrapper: CodeWrapper, test_list: CheckFunctionDict
):
    code = function_wrapper(b"")
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 2



from typing import Optional

from tree_sitter import Node

from castep_linter.fortran.fortran_nodes.argument_types import (
    ArgumentItem,
    KeywordArgument,
    PositionalArgument,
)
from castep_linter.fortran.fortran_nodes.fortran_node import FortranNode
from castep_linter.fortran.fortran_raw_types import Fortran
from castep_linter.fortran.identifier import Identifier


class FortranArgumentList(FortranNode):
    """Parser for fortran argument lists"""

    def __init__(self, arg_list: Optional[Node]):
        if arg_list:
            super().__init__(arg_list)

        if arg_list:
            self.args, self.kwargs = self._parse_arg_list()
        else:
            self.args, self.kwargs = [], {}

    def get_arg(self, keyword: Identifier, position: Optional[int] = None) -> ArgumentItem:
        """Return a value from a fortran argument list by keyword and optionally position"""
        if position and len(self.args) >= position:
            return PositionalArgument(self.args[position - 1])
        if keyword in self.kwargs:
            return KeywordArgument(self.kwargs[keyword])

        err = f"Argument {keyword} not found in argument list"
        raise KeyError(err)

    def _parse_arg_list(self):
        """
        Convert a fortran argument list into a args, kwargs pair.
        The keyword arguments will be converted into identifiers.
        """

        args = []
        kwargs = {}

        parsing_arg_list = True

        for child in self.children[1:-1]:
            if child.raw == ",":
                continue

            if child.is_type(Fortran.KEYWORD_ARGUMENT):
                parsing_arg_list = False

            if parsing_arg_list:
                args.append(child)
            elif child.is_type(Fortran.KEYWORD_ARGUMENT):
                key, _, value = child.children
                kwargs[Identifier.from_node(key)] = value
            elif child.is_type(Fortran.COMMENT):
                continue
            else:
                err = f"Unknown argument list item in keyword arguments: {child.type}: \n{child.raw}\nin\n{self.raw}"
                raise ValueError(err)

        return args, kwargs



"""Module to handle errors, warnings and info messages"""

from enum import Enum, auto
from typing import ClassVar, Dict, Literal

from castep_linter.fortran.fortran_nodes import FortranNode


class PrintStyle(Enum):
    """Error print styles"""

    ANNOTATED = auto()
    GCC = auto()


class FortranMsgBase:
    """Base class for other static analysis problems to inherit from"""

    ERROR_TYPE: ClassVar[str] = "NONE"
    ERROR_STYLE: ClassVar[str] = "grey"
    LINE_NUMBER_OFFSET = 8
    ERROR_SEVERITY: ClassVar[int] = 100

    def __init__(self, node: FortranNode, message: str) -> None:
        self.message = message
        self.start_point = node.node.start_point  # TODO FIX
        self.end_point = node.node.end_point

    def print_err(
        self, filename: str, console, *, print_style: PrintStyle = PrintStyle.ANNOTATED
    ) -> None:
        """Print the error to the supplied console"""

        if print_style is PrintStyle.ANNOTATED:
            console.print(self, style=self.ERROR_STYLE)
            context = self.context(filename, underline=True)
        elif print_style is PrintStyle.GCC:
            context = self._gcc_format(filename)

        if context:
            console.print(context)

    def _gcc_format(self, filename):
        """Format errors like the GCC"""
        start_line, _ = self.line_ranges
        start_char, _ = self.char_ranges

        return f"{filename}:{start_line+1}:{start_char}: {self.ERROR_TYPE}: {self.message}"

    def context(self, filename, *, underline=False):
        """Print a line of context for the current error"""
        context = ""

        with open(filename, "rb") as fd:
            start_line, _ = self.line_ranges
            start_char, _ = self.char_ranges

            file_str = str(filename)

            if self.num_lines == 1:
                line = fd.read().splitlines()[start_line].decode(errors="replace")
                context = f"{file_str}:{start_line+1:{self.LINE_NUMBER_OFFSET}}>{line}"
                if underline:
                    context += (
                        "\n"
                        + " " * (len(file_str) + 1)
                        + " " * (self.LINE_NUMBER_OFFSET + 1)
                        + " " * start_char
                        + "^" * self.num_chars
                    )
        return context

    @property
    def line_ranges(self):
        """Start and end line as a tuple"""
        return self.start_point[0], self.end_point[0]

    @property
    def num_lines(self):
        """Number of lines covering issue"""
        return self.line_ranges[1] - self.line_ranges[0] + 1

    @property
    def char_ranges(self):
        """Start and end chars as a tuple"""
        return self.start_point[1], self.end_point[1]

    @property
    def num_chars(self):
        """Number of characters covering issue"""
        return self.char_ranges[1] - self.char_ranges[0]

    def __repr__(self):
        return f"{self.ERROR_TYPE}: {self.message}"


class FortranError(FortranMsgBase):
    """Fatal static analysis problem in code"""

    ERROR_TYPE: ClassVar[str] = "Error"
    ERROR_STYLE: ClassVar[str] = "red"
    ERROR_SEVERITY: ClassVar[int] = 2


class FortranWarning(FortranMsgBase):
    """Warning message from static analysis"""

    ERROR_TYPE: ClassVar[str] = "Warning"
    ERROR_STYLE: ClassVar[str] = "yellow"
    ERROR_SEVERITY: ClassVar[int] = 1


class FortranInfo(FortranMsgBase):
    """Warning message from static analysis"""

    ERROR_TYPE: ClassVar[str] = "Info"
    ERROR_STYLE: ClassVar[str] = "Blue"
    ERROR_SEVERITY: ClassVar[int] = 0


def new_fortran_error(level: str, node: FortranNode, message: str) -> FortranMsgBase:
    """Generate a new fortran diagnostic message"""
    cls = FortranMsgBase
    if level == "Error":
        cls = FortranError
    elif level == "Warning":
        cls = FortranWarning
    elif level == "Info":
        cls = FortranInfo
    else:
        raise ValueError("Unknown fortran diagnostic message type: " + level)
    return cls(node, message)


ErrorNames = Literal["Error", "Warn", "Info"]

FORTRAN_ERRORS: Dict[ErrorNames, type[FortranMsgBase]] = {
    "Error": FortranError,
    "Warn": FortranWarning,
    "Info": FortranInfo,
}

ERROR_SEVERITY: Dict[str, int] = {k: v.ERROR_SEVERITY for k, v in FORTRAN_ERRORS.items()}



"""Test that a subroutine or function has a trace_entry and trace_exit with the correct name"""

from castep_linter.error_logging import ErrorLogger
from castep_linter.fortran.fortran_nodes import (
    FortranCallExpression,
    FortranNode,
    FortranVariableDeclaration,
)
from castep_linter.fortran.fortran_raw_types import Fortran, FType
from castep_linter.fortran.identifier import Identifier
from castep_linter.fortran.node_type_err import WrongNodeError
from castep_linter.tests import castep_identifiers


def check_trace_entry_exit(node: FortranNode, error_log: ErrorLogger) -> None:
    """Test that a subroutine or function has a trace_entry and trace_exit with the correct name"""

    if node.is_type(Fortran.SUBROUTINE) or node.is_type(Fortran.FUNCTION):
        subroutine_name = node.get_context_identifier()
    else:
        err = "Wrong node type passed"
        raise WrongNodeError(err)

    has_trace_entry = False
    has_trace_exit = False

    const_string_vars: dict[Identifier, str] = {}

    for var_node in node.get_children_by_name(Fortran.VARIABLE_DECLARATION):
        if not isinstance(var_node, FortranVariableDeclaration):
            err = f"Expected variable declaration but got {node.type}"
            raise WrongNodeError(err)

        var_decl = var_node

        if var_decl.var_type != FType.CHARACTER:
            continue

        const_string_vars.update(var_decl.get_initialized_vars())

    for routine in node.get_children_by_name(Fortran.SUBROUTINE_CALL):
        if not isinstance(routine, FortranCallExpression):
            err = f"Expected subroutine call but got {node.type}"
            raise WrongNodeError(err)

        if routine.name == castep_identifiers.TRACE_ENTRY:
            has_trace_entry = True
        elif routine.name == castep_identifiers.TRACE_EXIT:
            has_trace_exit = True

        if routine.name not in [castep_identifiers.TRACE_ENTRY, castep_identifiers.TRACE_EXIT]:
            continue

        try:
            trace_node = routine.get_arg(position=1, keyword=castep_identifiers.TRACE_STRING).value
        except KeyError:
            err = f"Unparsable name passed to trace in {subroutine_name}"
            error_log.add_msg("Error", routine, err)
            continue

        if trace_node.is_type(Fortran.STRING_LITERAL):
            trace_string = trace_node.parse_string_literal().lower()
            if trace_string != subroutine_name:
                err = f"Incorrect name passed to trace in {subroutine_name}"
                error_log.add_msg("Error", trace_node, err)

        elif trace_node.is_type(Fortran.IDENTIFIER):
            trace_sub_text = Identifier.from_node(trace_node)

            if trace_sub_text in const_string_vars:
                trace_string = const_string_vars[trace_sub_text]

                if trace_string.lower() != subroutine_name:
                    err = (
                        f"Incorrect name passed to trace in {subroutine_name} "
                        f'by variable {trace_sub_text}="{trace_string}"'
                    )
                    error_log.add_msg("Error", trace_node, err)
            else:
                err = f"Unidentified variable {trace_sub_text} passed to trace in {subroutine_name}"
                error_log.add_msg("Error", trace_node, err)

        else:
            err = f"Unrecognisable {routine.raw} {trace_node.type=} {routine}"
            raise ValueError(err)

    if not has_trace_entry:
        error_log.add_msg("Info", node, f"Missing trace_entry in {subroutine_name}")
    if not has_trace_exit:
        error_log.add_msg("Info", node, f"Missing trace_exit in {subroutine_name}")



"""Tests for Fortran code in CASTEP"""

import pathlib
from importlib import resources as impresources
from typing import Callable, Generator, Optional

from tree_sitter import Language, Parser

from castep_linter.fortran import node_factory
from castep_linter.fortran.fortran_nodes import FortranNode


def get_fortran_parser():
    """Get a tree-sitter-fortran parser from src"""

    tree_sitter_src_ref = impresources.files("castep_linter") / "tree_sitter_fortran"
    with impresources.as_file(tree_sitter_src_ref) as tree_sitter_src:
        fortran_language = Language(str(tree_sitter_src / "fortran.so"), "fortran")

    parser = Parser()
    parser.set_language(fortran_language)
    return parser


class FortranTree:
    """Parsed fortran source code tree"""

    def __init__(self, raw_text: bytes, parser: Optional[Parser] = None):
        if parser is None:
            parser = get_fortran_parser()

        self.raw_text = raw_text
        self.tree = parser.parse(self.raw_text)

    @staticmethod
    def from_file(file: pathlib.Path):
        """Read from a file and return a AST"""
        with file.open("rb") as fd:
            raw_text = fd.read()
        return FortranTree(raw_text)

    def walk(self) -> Generator[FortranNode, None, None]:
        """Traverse a tree-sitter tree in a depth first search"""
        cursor = self.tree.walk()

        reached_root = False
        while not reached_root:
            yield node_factory.wrap_node(cursor.node)

            if cursor.goto_first_child():
                continue

            if cursor.goto_next_sibling():
                continue

            retracing = True
            while retracing:
                if not cursor.goto_parent():
                    retracing = False
                    reached_root = True

                if cursor.goto_next_sibling():
                    retracing = False

    def display(self, printfn: Callable):
        """Print the tree for the source file"""
        node_factory.wrap_node(self.tree.root_node).print_tree(printfn)



# pylint: disable=W0621,C0116,C0114,C0121
import pytest

from castep_linter.fortran.identifier import Identifier


def test_identifier_equals():
    assert Identifier("x") == Identifier("x")
    assert Identifier("X") == Identifier("x")
    assert Identifier("x") == Identifier("X")
    assert Identifier("X") == Identifier("X")


def test_identifier_not_equals():
    assert Identifier("x") != Identifier("y")
    assert Identifier("X") != Identifier("y")
    assert Identifier("x") != Identifier("Y")
    assert Identifier("X") != Identifier("Y")


def test_string_equals():
    assert Identifier("x") == "x"
    assert Identifier("X") == "x"
    assert Identifier("x") == "X"
    assert Identifier("X") == "X"


def test_string_not_equals():
    assert Identifier("x") != "y"
    assert Identifier("X") != "y"
    assert Identifier("x") != "Y"
    assert Identifier("X") != "Y"


def test_identifier_hash_eq():
    assert hash(Identifier("x")) == hash(Identifier("x"))
    assert hash(Identifier("X")) == hash(Identifier("x"))
    assert hash(Identifier("x")) == hash(Identifier("X"))
    assert hash(Identifier("X")) == hash(Identifier("X"))


def test_identifier_hash_not_equals():
    assert hash(Identifier("x")) != hash(Identifier("y"))
    assert hash(Identifier("X")) != hash(Identifier("y"))
    assert hash(Identifier("x")) != hash(Identifier("Y"))
    assert hash(Identifier("X")) != hash(Identifier("Y"))


def test_identifier_compare_none():
    assert Identifier("x") != None  # noqa: E711
    assert Identifier("X") != None  # noqa: E711
    assert None != Identifier("x")  # noqa: E711
    assert None != Identifier("X")  # noqa: E711


def test_identifier_compare_object():
    with pytest.raises(TypeError):
        assert Identifier("x") != object()



"""Module containing useful classes for parsing a fortran source tree from tree-sitter"""

from typing import Callable, List, Optional, Tuple

from rich.console import Console
from tree_sitter import Node

from castep_linter.fortran import node_factory
from castep_linter.fortran.fortran_raw_types import Fortran, FortranLookup
from castep_linter.fortran.identifier import Identifier
from castep_linter.fortran.node_type_err import FortranContextError, WrongNodeError


class FortranNode:
    """Wrapper for tree_sitter Node type to add extra functionality"""

    def __init__(self, node: Node):
        self.node = node

        self.type: Optional[str]

        if self.node.is_named:
            self.type = self.node.type
        else:
            self.type = None

    @property
    def ftype(self) -> Fortran:
        """Return the node type as member of the Fortran enum"""
        if self.type in FortranLookup:
            return FortranLookup[self.type]
        else:
            return Fortran.UNKNOWN

    def is_type(self, ftype: Fortran) -> bool:
        """Checks if a fortran node is of the supplied type"""
        return self.ftype == ftype

    @property
    def children(self) -> List["FortranNode"]:
        """Return all children of this node"""
        return [node_factory.wrap_node(c) for c in self.node.children]

    def next_named_sibling(self) -> Optional["FortranNode"]:
        """Return the next named sibling of the current node"""
        if self.node.next_named_sibling:
            return node_factory.wrap_node(self.node.next_named_sibling)
        else:
            return None

    def get(self, ftype: Fortran) -> "FortranNode":
        """Return the first child node with the requested type"""
        for c in self.node.named_children:
            if c.type == ftype.value:
                return node_factory.wrap_node(c)

        err = f'"{ftype}" not found in children of node {self.raw}'
        raise KeyError(err)

    def get_children_by_name(self, ftype: Fortran) -> List["FortranNode"]:
        """Return all the children with the requested type"""
        return [
            node_factory.wrap_node(c) for c in self.node.named_children if c.type == ftype.value
        ]

    def split(self) -> Tuple["FortranNode", "FortranNode"]:
        """Split a relational node with a left and right part into the two child nodes"""
        left = self.node.child_by_field_name("left")

        if left is None:
            err = f"Unable to find left part of node pair: {self.raw}"
            raise KeyError(err)

        right = self.node.child_by_field_name("right")

        if right is None:
            err = f"Unable to find right part of node pair: {self.raw}"
            raise KeyError(err)

        return node_factory.wrap_node(left), node_factory.wrap_node(right)

    @property
    def raw(self) -> str:
        """Return a string of all the text in a node as unicode"""
        return self.node.text.decode()

    def parse_string_literal(self) -> str:
        "Parse a string literal object to get the string"
        if not self.type == "string_literal":
            err = f"Tried to parse {self.raw} as string literal"
            raise WrongNodeError(err)
        return self.raw.strip("\"'")

    def print_tree(self, printfn: Optional[Callable] = None, indent: int = 0):
        """Prints a representation of the tree"""
        if not printfn:
            printfn = Console().print

        if self.node.is_named:
            printfn(" │ " * indent + " ├ " + self.node.type)
        else:
            printfn(" │ " * indent + " ├ " + "[blue]" + self.node.type + "[/blue]")

        for c in self.children:
            c.print_tree(printfn, indent + 1)

    def get_context_identifier(self) -> Identifier:
        """Get the name of the containing context of this node"""
        if not self.node.parent:
            err = "Node has no parent!"
            raise FortranContextError(err)

        p = node_factory.wrap_node(self.node.parent)
        return p.get_context_identifier()



# pylint: disable=W0621,C0116,C0114
from unittest import mock

import pytest

from castep_linter.fortran.node_type_err import WrongNodeError
from castep_linter.scan_files import run_tests_on_code
from castep_linter.tests import CheckFunctionDict, check_real_dp_declaration
from tests.conftest import Parser


@pytest.fixture
def test_list() -> CheckFunctionDict:
    return {"variable_declaration": [check_real_dp_declaration]}


def test_wrong_node():
    mock_node = mock.Mock(**{"is_type.return_value": False})
    err_log = mock.MagicMock()
    with pytest.raises(WrongNodeError):
        check_real_dp_declaration(mock_node, err_log)


def test_complex_dp_correct(parse: Parser, test_list: CheckFunctionDict):
    code = b"complex(kind=dp) :: y"
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 0


def test_complex_dp_no_kind(parse: Parser, test_list: CheckFunctionDict):
    code = b"complex, intent(in) :: y"
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 1


def test_complex_integer_kind_with_keyword(parse: Parser, test_list: CheckFunctionDict):
    code = b"complex(kind=8) :: y"
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 1


def test_complex_integer_kind_without_keyword(parse: Parser, test_list: CheckFunctionDict):
    code = b"complex(8) :: y"
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 1


def test_complex_bad_var_kind_without_keyword(parse: Parser, test_list: CheckFunctionDict):
    code = b"complex(x) :: y"
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 1


def test_complex_bad_var_kind_with_keyword(parse: Parser, test_list: CheckFunctionDict):
    code = b"complex(kind=x) :: y"
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 1



# pylint: disable=W0621,C0116,C0114
from unittest import mock

import pytest

from castep_linter.fortran.node_type_err import WrongNodeError
from castep_linter.scan_files import run_tests_on_code
from castep_linter.tests import CheckFunctionDict, check_allocate_has_stat
from tests.conftest import CodeWrapper, Parser


@pytest.fixture
def test_list() -> CheckFunctionDict:
    return {"call_expression": [check_allocate_has_stat]}


def test_wrong_node():
    mock_node = mock.Mock(**{"is_type.return_value": False})
    err_log = mock.MagicMock()
    with pytest.raises(WrongNodeError):
        check_allocate_has_stat(mock_node, err_log)


def test_allocate_stat_correct(
    parse: Parser, subroutine_wrapper: CodeWrapper, test_list: CheckFunctionDict
):
    code = subroutine_wrapper(
        b"""
    allocate(stat_checked_var(x,y,z), stat=u)
    if (u/=0) call io_abort("bleh")
    """
    )
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 0


def test_allocate_stat_correct_wrong_way(
    parse: Parser, subroutine_wrapper: CodeWrapper, test_list: CheckFunctionDict
):
    code = subroutine_wrapper(
        b"""
    allocate(stat_checked_var(x,y,z), stat=u)
    if (0/=u) call io_abort("bleh")
    """
    )
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 0


def test_allocate_stat_correct_if_but_not_checked(
    parse: Parser, subroutine_wrapper: CodeWrapper, test_list: CheckFunctionDict
):
    code = subroutine_wrapper(
        b"""
    allocate(stat_checked_var(x,y,z), stat=u)
    if (0/=z) call io_abort("bleh")
    """
    )
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 1


def test_allocate_stat_correct_mixed_caps(
    parse: Parser, subroutine_wrapper: CodeWrapper, test_list: CheckFunctionDict
):
    code = subroutine_wrapper(
        b"""
    allocate(stat_checked_var(x,y,z), stat=u)
    if (U/=0) call io_abort("bleh")
    """
    )
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 0, error_log.errors[0].message


def test_allocate_stat_correct_mixed_caps2(
    parse: Parser, subroutine_wrapper: CodeWrapper, test_list: CheckFunctionDict
):
    code = subroutine_wrapper(
        b"""
    allocate(stat_checked_var(x,y,z), stat=U)
    if (u/=0) call io_abort("bleh")
    """
    )
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 0, error_log.errors[0].message


def test_allocate_stat_correct_comment(
    parse: Parser, subroutine_wrapper: CodeWrapper, test_list: CheckFunctionDict
):
    code = subroutine_wrapper(
        b"""
    allocate(stat_checked_var(x,y,z), stat=u)
    ! comment
    if (u/=0) call io_abort("bleh")
    """
    )
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 0


def test_allocate_no_stat(
    parse: Parser, subroutine_wrapper: CodeWrapper, test_list: CheckFunctionDict
):
    code = subroutine_wrapper(
        b"""
    allocate(stat_checked_var(x,y,z))
    """
    )
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 1


def test_allocate_stat_not_checked(
    parse: Parser, subroutine_wrapper: CodeWrapper, test_list: CheckFunctionDict
):
    code = subroutine_wrapper(
        b"""
    allocate(stat_checked_var(x,y,z), stat=u)
    """
    )
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 1


def test_allocate_stat_not_checked_with_line_after(
    parse: Parser, subroutine_wrapper: CodeWrapper, test_list: CheckFunctionDict
):
    code = subroutine_wrapper(
        b"""
    allocate(stat_checked_var(x,y,z), stat=u)
    x = 5
    """
    )
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 1


@pytest.mark.skip()
def test_allocate_stat_wrong_function_called(
    parse: Parser, subroutine_wrapper: CodeWrapper, test_list: CheckFunctionDict
):
    code = subroutine_wrapper(
        b"""
    allocate(stat_checked_var(x,y,z), stat=u)
    if (u/=0) call trace_exit()
    """
    )
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 1


def test_allocate_abort_correct(
    parse: Parser, subroutine_wrapper: CodeWrapper, test_list: CheckFunctionDict
):
    code = subroutine_wrapper(
        b"""
    allocate(stat_checked_var(x,y,z), stat=u)
    if (u/=0) call io_allocate_abort("stat_checked_var", "x")
    """
    )
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 0


@pytest.mark.skip()
def test_allocate_abort_wrong_var_name(
    parse: Parser, subroutine_wrapper: CodeWrapper, test_list: CheckFunctionDict
):
    code = subroutine_wrapper(
        b"""
    allocate(stat_checked_var(x,y,z), stat=u)
    if (u/=0) call io_allocate_abort("wrong_var", "x")
    """
    )
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 1


@pytest.mark.skip()
def test_allocate_abort_wrong_sub_name(
    parse: Parser, subroutine_wrapper: CodeWrapper, test_list: CheckFunctionDict
):
    code = subroutine_wrapper(
        b"""
    allocate(stat_checked_var(x,y,z), stat=u)
    if (u/=0) call io_allocate_abort("stat_checked_var", "wrong_sub_name")
    """
    )
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 1


@pytest.mark.skip()
def test_allocate_abort_wrong_var_and_sub(
    parse: Parser, subroutine_wrapper: CodeWrapper, test_list: CheckFunctionDict
):
    code = subroutine_wrapper(
        b"""
    allocate(stat_checked_var(x,y,z), stat=u)
    if (u/=0) call io_allocate_abort("wrong", "wrong_sub_name")
    """
    )
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 2



from enum import Enum, auto


class Fortran(Enum):
    """Represents raw fortran source code tree elements"""

    COMMENT = "comment"
    SUBROUTINE = "subroutine"
    SUBROUTINE_STMT = "subroutine_statement"
    FUNCTION = "function"
    FUNCTION_STMT = "function_statement"
    NAME = "name"
    SIZE = "size"
    INTRINSIC_TYPE = "intrinsic_type"
    ASSIGNMENT_STMT = "assignment_statement"
    ARGUMENT_LIST = "argument_list"
    SUBROUTINE_CALL = "subroutine_call"
    IDENTIFIER = "identifier"
    VARIABLE_DECLARATION = "variable_declaration"
    RELATIONAL_EXPR = "relational_expression"
    IF_STMT = "if_statement"
    PAREN_EXPRESSION = "parenthesized_expression"
    KEYWORD_ARGUMENT = "keyword_argument"
    STRING_LITERAL = "string_literal"
    NUMBER_LITERAL = "number_literal"
    TYPE_QUALIFIER = "type_qualifier"
    CALL_EXPRESSION = "call_expression"

    UNKNOWN = "unknown"


class FType(Enum):
    """Intrinsic variable types in fortran"""

    REAL = auto()
    DOUBLE = auto()
    COMPLEX = auto()
    INTEGER = auto()
    LOGICAL = auto()
    CHARACTER = auto()
    OTHER = auto()


FortranLookup = {k.value: k for k in Fortran}
FortranContexts = {Fortran.SUBROUTINE, Fortran.FUNCTION}



"""Module representing parsed fortran nodes"""

from castep_linter.fortran.fortran_nodes.fortran_argument_list import FortranArgumentList
from castep_linter.fortran.fortran_nodes.fortran_call_expression_node import FortranCallExpression
from castep_linter.fortran.fortran_nodes.fortran_function_node import FortranFunctionNode
from castep_linter.fortran.fortran_nodes.fortran_node import FortranNode
from castep_linter.fortran.fortran_nodes.fortran_subroutine_node import FortranSubroutineNode
from castep_linter.fortran.fortran_nodes.fortran_var_decl_node import FortranVariableDeclaration

__all__ = [
    "FortranNode",
    "FortranArgumentList",
    "FortranCallExpression",
    "FortranFunctionNode",
    "FortranSubroutineNode",
    "FortranVariableDeclaration",
]



"""Module holding methods for type checking tree_sitter Node functions"""


class WrongNodeError(Exception):
    """Exception thrown when an invalid node is passed to a typed function"""


class FortranContextError(Exception):
    """Exception thrown when some fortran code does not have a containing context (eg subroutine or function)"""



""""Tests to be performed by the CASTEP Fortran linter"""

from typing import Callable

from castep_linter.error_logging.logger import ErrorLogger
from castep_linter.fortran.fortran_nodes import FortranNode
from castep_linter.tests.allocate_stat_checked import check_allocate_has_stat
from castep_linter.tests.complex_has_dp import check_complex_has_dp
from castep_linter.tests.has_trace_entry_exit import check_trace_entry_exit
from castep_linter.tests.number_literal_correct_kind import check_number_literal
from castep_linter.tests.real_declaration_has_dp import check_real_dp_declaration

CheckFunction = Callable[[FortranNode, ErrorLogger], None]
CheckFunctionDict = dict[str, list[CheckFunction]]

test_list: CheckFunctionDict = {
    "variable_declaration": [check_real_dp_declaration],
    "subroutine": [check_trace_entry_exit],
    "function": [check_trace_entry_exit],
    "call_expression": [check_complex_has_dp, check_allocate_has_stat],
    "number_literal": [check_number_literal],
}



from typing import Optional

from tree_sitter import Node

from castep_linter.fortran.fortran_nodes.argument_types import ArgumentItem
from castep_linter.fortran.fortran_nodes.fortran_argument_list import FortranArgumentList
from castep_linter.fortran.fortran_nodes.fortran_node import FortranNode
from castep_linter.fortran.fortran_raw_types import Fortran
from castep_linter.fortran.identifier import Identifier
from castep_linter.fortran.node_type_err import WrongNodeError


class FortranCallExpression(FortranNode):
    """Class representing a Fortran call expression"""

    def __init__(self, call_expression_node: Node) -> None:
        super().__init__(call_expression_node)

        self.name = self._get_name()

        try:
            args = self.get(Fortran.ARGUMENT_LIST)
        except KeyError:
            args = FortranArgumentList(None)

        if not isinstance(args, FortranArgumentList):
            err = f"Expected argument list but got {args}"
            raise WrongNodeError(err)

        self.args = args

    def _get_name(self) -> Identifier:
        try:
            return Identifier.from_node(self.get(Fortran.IDENTIFIER))
        except KeyError:
            return Identifier("")

    def get_arg(self, keyword: Identifier, position: Optional[int] = None) -> ArgumentItem:
        """Get an argument from the call expression"""
        return self.args.get_arg(keyword, position)

    def __str__(self):
        return f"{self.name=} {self.args=}"



"""Module to write code linting errors in Jenkins json format"""

import json
from pathlib import Path
from typing import Dict, List, Literal, TypedDict

from castep_linter.error_logging.error_types import FORTRAN_ERRORS
from castep_linter.error_logging.logger import ErrorLogger

JSONSeverityLevel = Literal["LOW", "NORMAL", "HIGH", "ERROR"]

JSON_severity_dict: Dict[int, JSONSeverityLevel] = {
    FORTRAN_ERRORS["Error"].ERROR_SEVERITY: "HIGH",
    FORTRAN_ERRORS["Warn"].ERROR_SEVERITY: "NORMAL",
    FORTRAN_ERRORS["Info"].ERROR_SEVERITY: "LOW",
}


class JSONIssue(TypedDict):
    """Represents a single issue in a Jenkins JSON report"""

    fileName: str
    severity: JSONSeverityLevel
    message: str
    type: str
    lineStart: int
    lineEnd: int
    columnStart: int
    columnEnd: int


class JSONReport(TypedDict):
    """Represents an entire report"""

    _class: str
    issues: List[JSONIssue]
    size: int


def write_json(file: Path, error_logs: Dict[str, ErrorLogger], error_level: int):
    """write code linting errors in Jenkins json format"""

    issues: List[JSONIssue] = [
        JSONIssue(
            fileName=scanned_file,
            severity=JSON_severity_dict[error.ERROR_SEVERITY],
            message=error.message,
            type=determine_type(error.message),
            lineStart=error.start_point[0] + 1,  # Jenkins lines 1-indexed
            lineEnd=error.end_point[0] + 1,
            columnStart=error.start_point[1] + 1,
            columnEnd=error.end_point[1] + 1,
        )
        for scanned_file, log in error_logs.items()
        for error in log.errors
        if error.ERROR_SEVERITY >= error_level
    ]

    report: JSONReport = {
        "_class": "io.jenkins.plugins.analysis.core.restapi.ReportApi",
        "issues": issues,
        "size": len(issues),
    }

    with open(file, "w", encoding="utf-8") as out_file:
        json.dump(report, out_file, indent=2)


def determine_type(message: str) -> str:
    """Determine type of error from key components"""
    if "alloc" in message.lower():
        return "ALLOC"

    if "complex intrinsic" in message:
        return "CMPLX_KIND"

    if "literal" in message:
        return "LITERAL_KIND"

    if "kind" in message or "Kind" in message:
        return "KIND"

    if "Missing trace_" in message or "Incorrect name" in message:
        return "TRACE"

    return "UNKNOWN"



"""Commonly used identifiers in CASTEP"""

from castep_linter.fortran.identifier import Identifier

# TRACE THINGS
TRACE_ENTRY = Identifier("trace_entry")
TRACE_EXIT = Identifier("trace_exit")
TRACE_STRING = Identifier("string")


CMPLX = Identifier("cmplx")

# Parameters
DP = Identifier("dp")
DPREC = Identifier("dprec")
DI_DP = Identifier("di_dp")
VERSION_KIND = Identifier("version_kind")

DP_ALL = {DP, DPREC, DI_DP, VERSION_KIND}

# Integer kinds
INT32 = Identifier("int32")
INT64 = Identifier("int64")
INT_KINDS = {INT32, INT64}

# Special keywords
STAT = Identifier("stat")
KIND = Identifier("kind")
ALLOCATE = Identifier("allocate")

IO_ABORT = Identifier("io_abort")
IO_ALLOCATE_ABORT = Identifier("io_allocate_abort")

IO_ABORT_ANY = {IO_ABORT, IO_ALLOCATE_ABORT}

ARRAY = Identifier("---NONE---")


IO_AA_ARRAY = Identifier("array_name")
IO_AA_ROUTINE = Identifier("routine_name")



"""Test that a number literal has a dp (if real) or no dp if of any other type"""

from castep_linter.error_logging import ErrorLogger
from castep_linter.fortran.fortran_nodes import FortranNode
from castep_linter.fortran.fortran_raw_types import Fortran
from castep_linter.fortran.node_type_err import WrongNodeError
from castep_linter.tests import castep_identifiers


def check_number_literal(node: FortranNode, error_log: ErrorLogger) -> None:
    """Test that a number literal has a dp (if real) or no dp if of any other type"""

    if not node.is_type(Fortran.NUMBER_LITERAL):
        err = "Expected number literal node"
        raise WrongNodeError(err)

    literal_string = node.raw.lower()

    if "_" in literal_string:
        value, kind = literal_string.split("_", maxsplit=1)

        if is_int(value):
            if kind not in castep_identifiers.INT_KINDS:
                error_log.add_msg("Error", node, f"Integer literal with {kind=}")
        elif kind not in castep_identifiers.DP_ALL:
            error_log.add_msg("Error", node, f"Float literal with {kind=}")

    elif "d" in literal_string:
        pass  # eg 5.0d4
    elif not is_int(literal_string):
        error_log.add_msg("Error", node, "Float literal without kind")


def is_int(x: str) -> bool:
    return not any(c in x.lower() for c in [".", "e", "d"])



"""Static code analysis tool for castep"""

import argparse
import logging
import pathlib
import sys

from rich.console import Console

from castep_linter import error_logging
from castep_linter.error_logging.error_types import PrintStyle
from castep_linter.error_logging.json_writer import write_json
from castep_linter.error_logging.xml_writer import write_xml
from castep_linter.fortran import parser
from castep_linter.tests import CheckFunction, test_list

# done - complex(var) vs complex(var,dp) or complex(var, kind=dp)
# done - allocate without stat and stat not checked. deallocate?
# done - integer_dp etc
# real with trailing . not .0 or .0_dp?
# io_allocate_abort with wrong subname
# tabs & DOS line endings, whitespace, comments?


def run_tests_on_code(
    fort_tree: parser.FortranTree, test_dict: dict[str, list[CheckFunction]], filename: str
) -> error_logging.ErrorLogger:
    """Run all available tests on the supplied source code"""
    error_log = error_logging.ErrorLogger(filename)

    for node in fort_tree.walk():
        # Have to check for is_named here as we want the statements,
        # not literal words like subroutine
        if node.type in test_dict:
            for test in test_dict[node.type]:
                test(node, error_log)

    return error_log


def path(arg: str) -> pathlib.Path:
    """Check a file exists and if so, return a path object"""
    my_file = pathlib.Path(arg)
    if not my_file.is_file():
        err = f"The file {arg} does not exist!"
        raise argparse.ArgumentTypeError(err)
    return my_file


def parse_args():
    """Parse the command line args for a message print level and a list of filenames"""
    arg_parser = argparse.ArgumentParser(prog="castep-linter", description="Code linter for CASTEP")
    arg_parser.add_argument(
        "-l",
        "--level",
        help="Error message level",
        default="Info",
        choices=error_logging.ERROR_SEVERITY.keys(),
    )
    arg_parser.add_argument(
        "-f",
        "--format",
        type=lambda x: str(x).upper(),
        choices=[x.name for x in PrintStyle],
        help="Format screen printing",
        default="ANNOTATED",
    )
    arg_parser.add_argument(
        "-x", "--xml", type=pathlib.Path, help="File for JUnit xml output if required"
    )
    arg_parser.add_argument(
        "-j", "--json", type=pathlib.Path, help="File for Jenkins json output if required"
    )
    arg_parser.add_argument("-q", "--quiet", action="store_true", help="Do not write to console")
    arg_parser.add_argument("-d", "--debug", action="store_true", help="Turn on debug output")
    arg_parser.add_argument(
        "-p", "--print-tree", action="store_true", help="Print the parsed source tree"
    )
    arg_parser.add_argument("file", nargs="+", type=path, help="Files to scan")
    return arg_parser.parse_args()


def main() -> None:
    """Main entry point for the CASTEP linter"""
    args = parse_args()

    console = Console(soft_wrap=True)

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)

    error_logs = {}

    for file in args.file:
        # Parse the source file
        fortan_tree = parser.FortranTree.from_file(file)

        # Print for development
        if args.print_tree:
            fortan_tree.display(console.print)

        # Actually run the tests
        try:
            error_log = run_tests_on_code(fortan_tree, test_list, str(file))
        except UnicodeDecodeError:
            logging.error("Failed to properly decode %s", file)
            raise
        except Exception:
            logging.error("Failed to properly parse %s", file)
            raise

        # Report any errors
        if not args.quiet:
            error_log.print_errors(console, level=args.level, print_style=PrintStyle[args.format])

            err_count = error_log.count_errors()

            console.print(
                f"{len(error_log.errors)} issues in {file} ({err_count['Error']} errors,"
                f" {err_count['Warn']} warnings, {err_count['Info']} info)"
            )

        error_logs[str(file)] = error_log

    # Write junit xml file
    if args.xml:
        write_xml(args.xml, error_logs, error_logging.ERROR_SEVERITY[args.level])
    if args.json:
        write_json(args.json, error_logs, error_logging.ERROR_SEVERITY[args.level])

    # Exit with an error code if there were any errors
    if any(e.has_errors_above(args.level) for e in error_logs.values()):
        sys.exit(1)
    else:
        sys.exit(0)



from importlib.resources import as_file, files

from tree_sitter import Language

file_ref = files("castep_linter") / "tree_sitter_fortran"
with as_file(file_ref) as tree_sitter_dir:
    Language.build_library(str(tree_sitter_dir / "fortran.so"), [str(tree_sitter_dir)])



# pylint: disable=W0621,C0116,C0114
import pytest

from castep_linter.scan_files import run_tests_on_code
from castep_linter.tests import CheckFunctionDict, check_complex_has_dp
from tests.conftest import Parser


@pytest.fixture
def test_list() -> CheckFunctionDict:
    return {"call_expression": [check_complex_has_dp]}


def test_other_function(parse: Parser, test_list: CheckFunctionDict):
    code = b"y = myownfunction(a, b, dp)"
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 0


def test_complex_dp_correct(parse: Parser, test_list: CheckFunctionDict):
    code = b"y = CMPLX(a, b, dp)"
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 0


def test_complex_dp_correct_keyword(parse: Parser, test_list: CheckFunctionDict):
    code = b"y = CMPLX(z, kind=dp)"
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 0


def test_complex_dp_wrong_place(parse: Parser, test_list: CheckFunctionDict):
    code = b"y = CMPLX(z, dp)"
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 1


def test_complex_dp_missing(parse: Parser, test_list: CheckFunctionDict):
    code = b"y = CMPLX(a, b)"
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 1


def test_complex_wrong_kind(parse: Parser, test_list: CheckFunctionDict):
    code = b"y = CMPLX(a, b, x)"
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 1



"""Module containing code for the high level error logger"""

from collections import Counter
from dataclasses import dataclass, field
from typing import Iterator, List

from rich.console import Console

from castep_linter.error_logging import error_types
from castep_linter.fortran.fortran_nodes import FortranNode


@dataclass
class ErrorLogger:
    """Container for all errors and messages generated while analysing
    a Fortran file"""

    filename: str
    errors: List[error_types.FortranMsgBase] = field(default_factory=list)

    def __iter__(self) -> Iterator[error_types.FortranMsgBase]:
        return iter(self.errors)

    def add_msg(self, level: str, node: FortranNode, message: str):
        """Add an error to the error list"""
        err = error_types.new_fortran_error(level, node, message)
        self.errors.append(err)

    def print_errors(
        self,
        console: Console,
        level: str = "Warning",
        *,
        print_style: error_types.PrintStyle = error_types.PrintStyle.ANNOTATED,
    ) -> None:
        """Print all the contained errors above the level"""
        severity = error_types.ERROR_SEVERITY[level]

        for err in self.errors:
            if err.ERROR_SEVERITY >= severity:
                err.print_err(self.filename, console, print_style=print_style)

    def count_errors(self):
        """Count the number of errors in each category"""
        c = Counter(e.ERROR_SEVERITY for e in self.errors)
        return {
            err_str: c[err_severity] for err_str, err_severity in error_types.ERROR_SEVERITY.items()
        }

    def __len__(self):
        return len(self.errors)

    @property
    def has_errors(self):
        """Does the logger contain any errors"""
        return len(self) > 0

    def has_errors_above(self, level: str):
        """Does the logger contain any errors above the requested level"""
        error_severity = error_types.ERROR_SEVERITY[level]
        errors = (1 for e in self.errors if e.ERROR_SEVERITY >= error_severity)
        return sum(errors) > 0  # type: ignore



# pylint: disable=W0621,C0116,C0114
import pytest

from castep_linter.scan_files import run_tests_on_code
from castep_linter.tests import CheckFunctionDict, check_number_literal
from tests.conftest import Parser


@pytest.fixture
def test_list() -> CheckFunctionDict:
    return {"number_literal": [check_number_literal]}


def test_real_literal_with_dp(parse: Parser, test_list: CheckFunctionDict):
    code = b"z = 1.0_dp"
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 0


def test_real_literal_with_version_kind(parse: Parser, test_list: CheckFunctionDict):
    code = b"z = 1.0_version_kind"
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 0


def test_real_literal_with_other_kind(parse: Parser, test_list: CheckFunctionDict):
    code = b"z = 1.0_sp"
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 1


def test_real_literal_missing_dp(parse: Parser, test_list: CheckFunctionDict):
    code = b"z = 1.0"
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 1


def test_real_literal_missing_dp_scientific(parse: Parser, test_list: CheckFunctionDict):
    code = b"z = 1.0e5"
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 1


def test_real_literal_d0(parse: Parser, test_list: CheckFunctionDict):
    code = b"z = 1.0d5"
    error_log = run_tests_on_code(parse(code), test_list, "filename")
    assert len(error_log.errors) == 0



from tree_sitter import Node

from castep_linter.fortran import fortran_nodes
from castep_linter.fortran.fortran_raw_types import Fortran, FortranLookup


def wrap_node(node: Node) -> "fortran_nodes.FortranNode":
    """Turn a tree-sitter node into a FortranNode of the correct type"""

    factory_dict = {
        Fortran.SUBROUTINE: fortran_nodes.FortranSubroutineNode,
        Fortran.FUNCTION: fortran_nodes.FortranFunctionNode,
        Fortran.SUBROUTINE_CALL: fortran_nodes.FortranCallExpression,
        Fortran.CALL_EXPRESSION: fortran_nodes.FortranCallExpression,
        Fortran.ARGUMENT_LIST: fortran_nodes.FortranArgumentList,
        Fortran.VARIABLE_DECLARATION: fortran_nodes.FortranVariableDeclaration,
    }

    if node.is_named:
        node_type_raw = node.type
        node_type = FortranLookup.get(node_type_raw, Fortran.UNKNOWN)
    else:
        node_type = Fortran.UNKNOWN

    factory_method = factory_dict.get(node_type, fortran_nodes.FortranNode)
    return factory_method(node)



"""Module to write code linting errors in JUnit XML format"""

from pathlib import Path
from typing import Dict

from junitparser import Error, JUnitXml, Skipped, TestCase, TestSuite  # type: ignore

from castep_linter.error_logging.logger import ErrorLogger


def write_xml(file: Path, error_logs: Dict[str, ErrorLogger], error_level: int):
    """write code linting errors in JUnit XML format"""
    xml = JUnitXml()
    for scanned_file, log in error_logs.items():
        suite = TestSuite(scanned_file)

        for error in log.errors:
            case = TestCase(str(error))
            if error.ERROR_SEVERITY >= error_level:
                case.result = [Error(error.context(scanned_file, underline=True))]
            else:
                case.result = [Skipped(error.context(scanned_file, underline=True))]
            suite.add_testcase(case)

        xml.add_testsuite(suite)

    xml.write(str(file))



from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from castep_linter.fortran.fortran_nodes import FortranNode


@dataclass
class ArgumentItem:
    """Item in argument list"""

    value: "FortranNode"


@dataclass
class KeywordArgument(ArgumentItem):
    """Item in argument list specified by keyword"""


@dataclass
class PositionalArgument(ArgumentItem):
    """Item in argument list specified by position"""
